#!/usr/bin/perl -w
#
# ctlinetd - Control script for inetd.
# 
# Copyright (c) 1999-2007 Occam's Razor. All rights reserved.
#
# See the LICENSE file distributed with this code for restrictions on its use
# and further distribution.
# Original distribution available at <http://www.occam.com/tools/>.
#
# $Id: ctlinetd.pl,v 6.3 2007/03/27 13:43:51 leonvs Exp $
#
# TODO
#	listenabled command w/ xinetd
#		eliminate dupes from service list
#		don't list service if disabled in xinetd.conf defaults section
#	ability to specify tcp/udp versions of services
#		easy enough to do, but would it be useful?
#	fuller support for more OSes
#

use strict;
use File::Basename;
use Fcntl;

# global variables
my $DEBUG = 0;
my $progname;
my $VERSION;
my $USAGE;
my $os;
my $inetd;
my @configFiles = ();

setup();
processArgs();

sub setup {
	$progname = $0; $progname =~ s/.*\///;

	$VERSION = "$progname 6.3\n";

	$USAGE = <<EOF;
$VERSION
usage:	$progname help
		Prints this usage statement.

	$progname version
		Prints version of $progname.

	$progname status [service...]
		Confirms that inetd is running, and reports its process ID.
		If services are specified, reports on enabled/disabled status.

	$progname listenabled
		Reports on the status of all enabled services.

	$progname reload
		Sends a signal to inetd to reload its configuration.

	$progname restart
		Stops and starts inetd.

	$progname start
		Starts inetd if it isn't running.

	$progname stop
		Stops inetd.

	$progname enable service
		Enables the service, and reloads the config if necessary.

	$progname disable service
		Disables the service, and reloads the config if necessary.

	$progname commentall
		Disables all services. With traditional inetd or xinetd, the
		config is not reloaded, which is useful for getting the config
		files into an initial secure state. With other mechanisms
		(e.g., SRC, SMF, launchd), services are disabled immediately.

	$progname undo
		Replaces the config files with the backups, and reloads.
		USE WITH CAUTION.

EOF

	$os = $^O;
	if ($os eq "solaris") {
		my $solRev = `/sbin/uname -r`;
		chomp $solRev;
		report("DEBUG", "Solaris version is $solRev.");
		$os = "solaris10" if ($solRev eq "5.10");
	} elsif ($os eq "darwin") {
		my $darRev = `/usr/bin/uname -r`;
		chomp $darRev;
		report("DEBUG", "Darwin version is $darRev.");
		$os = "darwin8" if ($darRev =~ /^8/);
	}

	setInetdType();
}

sub processArgs {
	my $command;
	my $service;
	my @services;

	$command = shift @ARGV or die $USAGE;
	if ($command eq "status") {
		@services = @ARGV;
		@ARGV = ();
	} elsif ($command eq "enable" or $command eq "disable") {
		$service = shift @ARGV or die $USAGE;
	}

	# Should have no more args left.
	@ARGV and die $USAGE;

	for ($command) {
		/^help$/	and die $USAGE;
		/^version$/	and die $VERSION;
		/^status$/	and do { status(@services); last; };
		/^listenabled$/	and do { status("ENABLED"); last; };
		/^reload$/	and do { reload(); last; };
		/^restart$/	and do { restart(); last; };
		/^start$/	and do { start(); last; };
		/^stop$/	and do { stop(); last; };
		/^enable$/	and do { change($service, "ENABLE"); last; };
		/^disable$/	and do { change($service, "DISABLE"); last; };
		/^commentall$/	and do { change("", "DISABLE"); last; };
		/^undo$/	and do { undo(); last; };
		die $USAGE;
	}
}

sub status {
	my ($status, $IP6, $WRAPPED);

	foreach my $service (@_) {
		$service or next;	# undo() passes empty service to reload

		# If service is "ENABLED", add all enabled services to list.
		report("DEBUG", "inetd type is $inetd.");
		if ($inetd eq "src") {
			$service eq "ENABLED" and
				push @_, getEnabled_src() and next;
			($status) = getStatus_src($service);
		} elsif ($inetd eq "smf") {
			$service eq "ENABLED" and
				push @_, getEnabled_smf() and next;
			($status, $WRAPPED) = getStatus_smf($service);
		} elsif ($inetd eq "launchd") {
			$service eq "ENABLED" and
				push @_, getEnabled_launchd() and next;
			($status) = getStatus_launchd($service);
		} elsif ($inetd eq "xinetd") {
			$service eq "ENABLED" and
				push @_, getEnabled_x() and next;
			($status, $IP6) = getStatus_x($service);
		} else {
			$service eq "ENABLED" and
				push @_, getEnabled() and next;
			($status, $IP6, $WRAPPED) = getStatus($service);
		}

		print "$service is $status";
		print ", using IPv6" if $IP6;
		print ", using tcp_wrappers" if $WRAPPED;
		print "\n";
	}

	my @pids = getPIDs("(rpc\\.)?$inetd");
	my $pids = join " ", @pids;
	my $numProcs = ($os eq "dynixptx") ? 2 : 1;	# no. of expected procs
	my $overProcs = @pids - $numProcs;		# no. of "excess" procs
	my $process = ($numProcs > 1) ? "processes" : "process";

	if ($overProcs < 0) {
		print "inetd $process not running";
		if ($inetd eq "launchd") {
			my ($xinetdStatus) = getStatus_launchd("xinetd");
			($xinetdStatus eq "enabled") and
				print ", will run on demand";
		}
		print "\n";
	} elsif ($overProcs > 0) {
		print "Too many inetd processes running (pid $pids)\n";
	} else { #($overProcs == 0)
		print "inetd $process running (pid $pids)\n";
	}
	return $overProcs;
}

sub reload {
	my $service = shift;
	my @pids = getPIDs("(rpc\\.)?$inetd");
	my $sig = ($inetd eq "xinetd") ? "USR2" : "HUP";

	if (status($service) == 0) {
		if ($inetd eq "src") {
			system "/bin/refresh -s inetd";
		} elsif ($inetd eq "smf") {
			system "/usr/sbin/svcadm refresh inetd";
		} elsif ($inetd eq "launchd") {
			report("INFO", "xinetd reloaded on demand.");
		} else {
			kill $sig, @pids;
		}
		print "inetd reloaded\n";
		status();
	}
}

sub restart {
	if ($inetd eq "smf") {
		status();
		system "/usr/sbin/svcadm restart inetd";
		print "inetd restarted\n";
		sleep 3;	# give SMF time to work
		status();
	} elsif ($inetd eq "launchd") {
		report("INFO", "xinetd restarted on demand.");
		status();
	} else {
		stop();
		start();
	}
}

sub start {
	if (status() < 0) {
		my $xinetd = (-x "/usr/sbin/xinetd") ?
			"/usr/sbin/xinetd" : "/usr/local/sbin/xinetd";

		if ($inetd eq "src") {
			system "/bin/startsrc -s inetd";
		} elsif ($inetd eq "smf") {
			system "/usr/sbin/svcadm enable inetd";
		} elsif ($inetd eq "launchd") {
			report("INFO", "xinetd started on demand.");
		} elsif (-x $xinetd) {	# assume xinetd preferred if it exists
			$inetd = "xinetd";
			system "$xinetd -reuse -pidfile /var/run/xinetd.pid";
		} else {
			for ($os) {
				/^solaris$/ and do {
					system "/usr/sbin/inetd -s -t";
					last;
				};
				/^hpux$/ and do {
					system "/usr/sbin/inetd -l";
					last;
				};
				/^dynixptx$/ and do {
					system "/usr/sbin/inetd -v";
					last;
				};
				# Linux, Darwin/Mac OS X, ...
				# Also try this if the OS is unsupported.
				system "/usr/sbin/inetd";
			}
		}
		print "inetd started\n";
		status();
	}
}

sub stop {
	my @pids = getPIDs("(rpc\\.)?$inetd");

	if (status() == 0) {
		if ($inetd eq "src") {
			system "/bin/stopsrc -s inetd";
		} elsif ($inetd eq "smf") {
			system "/usr/sbin/svcadm disable inetd";
			sleep 3;	# give SMF time to work
		} elsif ($inetd eq "launchd") {
			report("INFO", "xinetd stopped on demand.");
		} else {
			kill "TERM", @pids;
		}
		print "inetd stopped\n";
		status();
	}
}

sub change {
	my $service = shift;
	my $changeFlag = shift;
	my ($status, $IP6, $WRAPPED, $defaultConfigFile, @inFiles);
	my $toStatus = ($changeFlag eq "DISABLE") ? "disabled" : "enabled";
	my $fromStatus = ($changeFlag eq "DISABLE") ? "enabled" : "disabled";
	my ($inPattern, $outPattern);
	my $inFile;

	if ($service) {			# null $service means all
		if ($inetd eq "src") {
			($status)
				= getStatus_src($service);
		} elsif ($inetd eq "smf") {
			($status, $WRAPPED)
				= getStatus_smf($service);
		} elsif ($inetd eq "launchd") {
			($status)
				= getStatus_launchd($service);
		} elsif ($inetd eq "xinetd") {
			($status, $IP6, $defaultConfigFile, @inFiles)
				= getStatus_x($service);
		} else {
			($status, $IP6, $WRAPPED, @inFiles)
				= getStatus($service);
		}
		$status =~ s/$toStatus/already $toStatus/;
		$status =~ /\b$fromStatus\b/
			or report("ERR", "$service is $status");

		if ($inetd eq "src") {
			if ($changeFlag eq "DISABLE") {
				system "/bin/stopsrc -t $service";
			} elsif ($changeFlag eq "ENABLE") {
				system "/bin/startsrc -t $service";
			} else {
				report("ERR", "Invalid change flag: $changeFlag");
			}
			status($service);
		} elsif ($inetd eq "smf") {
			my $flag = ($changeFlag eq "DISABLE") ? "-d" : "-e";
			system "/usr/sbin/inetadm $flag $service";
			sleep 3;	# give SMF time to work
			status($service);
		} elsif ($inetd eq "launchd") {
			my $act = ($changeFlag eq "DISABLE") ? "unload" : "load";
			my $svcFile = getConfig_launchd($service);
			report("DEBUG", "About to $act $svcFile.");
			system "/bin/launchctl $act -w $svcFile";
			status($service);
		} else {
			foreach $inFile (@inFiles) {
				if ($inetd eq "xinetd") {
					updateConfig_x($inFile, $defaultConfigFile, $service, $changeFlag);
				} else {
					if ($changeFlag eq "DISABLE") {
						$inPattern = "^[ \\t]*$service\\b";
						$outPattern = "#$service";
					} elsif ($changeFlag eq "ENABLE") {
						$inPattern = "^[ \\t]*#[# \\t]*$service\\b";
						$outPattern = "$service";
					} else {
						report("ERR", "Invalid change flag: $changeFlag");
					}
					updateConfig($inFile, $inPattern, $outPattern);
				}
			}
			reload($service);
		}
	} else {
		if ($inetd eq "xinetd") {
			# Just need the $defaultConfigFile.
			($status, $IP6, $defaultConfigFile, @inFiles)
				= getStatus_x($service);
			@configFiles = @inFiles;
		}

		# If all services specified, must be a disabling action.
		$changeFlag eq "DISABLE"
			or report("ERR", "Invalid change flag: $changeFlag");

		if ($inetd eq "src") {
			my @enabledSvcs = getEnabled_src();
			foreach my $svc (@enabledSvcs) {
				system "/bin/stopsrc -t $svc";
			}
		} elsif ($inetd eq "smf") {
			my @enabledSvcs = getEnabled_smf();
			foreach my $svc (@enabledSvcs) {
				system "/usr/sbin/inetadm -d $svc";
			}
		} elsif ($inetd eq "launchd") {
			my @enabledSvcs = getEnabled_launchd();
			foreach my $svc (@enabledSvcs) {
				my $svcFile = getConfig_launchd($svc);
				report("DEBUG", "About to unload $svcFile.");
				system "/bin/launchctl unload -w $svcFile";
			}
		} else {
			foreach $inFile (@configFiles) {
				if ($inetd eq "xinetd") {
					updateConfig_x($inFile, $defaultConfigFile, $service, $changeFlag);
				} else {
					# Replace at most one comment character,
					# and only leading whitespace.
					$inPattern = "^[ \\t]*#?";
					$outPattern = "#";
					updateConfig($inFile, $inPattern, $outPattern);
				}
			}
		}

		print "all services disabled\n";
	}
}

sub undo {
	if ($inetd eq "src") {
		report("INFO", "Undo not supported with AIX SRC.");
	} elsif ($inetd eq "smf") {
		report("INFO", "Undo not supported with SMF.");
	} elsif ($inetd eq "launchd") {
		report("INFO", "Undo not supported with launchd.");
	} else {
		foreach my $file (@configFiles) {
			my $dir = dirname $file;
			my $base = basename $file;
			my $tmpFile = "$dir/.$base.tmp";
			my $bakFile = "$dir/.$base.bak";

			$file or next;		# Don't do if filename unset.
			-f $bakFile or report("ERR", "No backup file $bakFile.");

			link $file, $tmpFile
				or report("ERR", "Can't link $tmpFile to $file.");
			rename $bakFile, $file
				or report("ERR", "Can't move $bakFile to $file.");
			rename $tmpFile, $bakFile
				or report("ERR", "Can't move $tmpFile to $bakFile.");
			print "Files $file and $bakFile swapped.\n";
		}
		reload();
	}
}

sub setInetdType {
	report("DEBUG", "OS is $os.");
	($os eq "aix") and $inetd = "src" and return;
	($os eq "solaris10") and $inetd = "smf" and return;
	($os eq "darwin8") and $inetd = "launchd" and return;

	# Make sure regular inetd isn't also somehow running.
	report("ERR", "Both inetd and xinetd are running!")
		if (getPIDs("inetd") && getPIDs("xinetd"));
	$inetd = getPIDs("xinetd") ? "xinetd" : "inetd";
	setConfigFiles();
}

# Returns list of PIDs for specified process.
sub getPIDs {
	my $procname = shift;
	my ($ps, $pidPos);
	my @pids;

	# Ugly hack, ruining the generality of this function, but it makes
	# things easier.
	($procname =~ /(src|smf)/) and $procname = "inetd";
	($procname =~ /launchd/) and $procname = "inetd";

	if ($os =~ /(solaris|dynixptx|hpux|irix|aix)/) {
		# SysV ps
		$ps = "/bin/ps -ef";
		$pidPos = 2;
	} else {
		# BSD ps
		$ps = "/bin/ps ax";
		$pidPos = 1;
	}

	# Look for process name preceded by slash or space, followed by space.
	my @procLine = `$ps`;
	@procLine = grep /(\/|\s)$procname\s/, @procLine;
	foreach (@procLine) {
		my @temp = split;
		push @pids, $temp[$pidPos - 1];
	}
	report("DEBUG", "PIDs are @pids.");

	return @pids;
}

sub setConfigFiles {
	my @configFilesToTry;
	my ($xinetdConfigFile, $dir);
	my ($file, $linkSrc, $path);

	if ($inetd eq "xinetd") {
		$xinetdConfigFile = (-e "/etc/xinetd.conf") ?
			"/etc/xinetd.conf" : "/usr/local/etc/xinetd.conf";
		push @configFilesToTry, "$xinetdConfigFile";

		# The following adds filenames specified in "include"
		# statements, and files in directories specified in "includedir"
		# statements. It will only process these statements in the main
		# file, not in any included files.
		open IN, "<$xinetdConfigFile"
			or report("ERR", "Can't open $xinetdConfigFile.");
		while (<IN>) {
			if (s/^[ \t]*include\b//) {
				push @configFilesToTry, split;
				next;
			}
			if (s/^[ \t]*includedir\b//) {
				foreach $dir (split) {
					opendir DIR, $dir;
					# Pick up all non-dot text files.
					push @configFilesToTry, grep -T,
						map "$dir/$_", grep !/^\./,
						readdir DIR;
					closedir DIR;
				}
				next;
			}
		}
		close IN;
	} else {
		push @configFilesToTry, "/etc/inetd.conf";
	}
	# Boy, Sequent makes this script *so* much more difficult.
	# Actually, it's nothing compared to what xinetd requires.
	push @configFilesToTry, "/etc/rpc.inetd.conf" if $os eq "dynixptx";

	while ($file = shift @configFilesToTry) {
		# Check if filename is a symlink. We want the actual filename.
		if ($linkSrc = readlink $file) {
			# If link is to relative pathname, tack on absolute path
			# and try again. Not pretty, but it works.
			if ($linkSrc !~ /^\//) {
				$path = $file; $path =~ s/\/[^\/]+$//;
				$linkSrc = "$path/$linkSrc";
				push @configFilesToTry, $linkSrc;
				next;
			}
			$file = $linkSrc;
		}
		push @configFiles, $file;
	}
}

# Returns list of enabled services.
sub getEnabled {
	my @enabled;

	foreach my $file (@configFiles) {
		open IN, "<$file" or report("ERR", "Can't open $file.");
		while (<IN>) {
			# Skip commented lines.
			/^[ \t]*#/ and next;
			my ($service) = split;
			push @enabled, $service if defined $service;
		}
		close IN;
	}

	return @enabled;
}

# Returns list of enabled services.
sub getEnabled_x {
	my $service;
	my ($IN_DEFAULTS_SECTION, $IN_SERVICE_SECTION);
	my %disabled;
	my $ENABLED_EXISTS;
	my %enabled;
	my @enabled;

	foreach my $file (@configFiles) {
		open IN, "<$file" or report("ERR", "Can't open $file.");
		while (<IN>) {
			if ($IN_DEFAULTS_SECTION) {
				if (/^[ \t]*disabled[ \t]*=[ \t]*(.+)/) {
					foreach my $service (split ' ', $+) {
						$disabled{$service}++;
					}
				}
				if (/^[ \t]*enabled[ \t]*=[ \t]*(.+)/) {
					$ENABLED_EXISTS++;
					foreach my $service (split ' ', $+) {
						$enabled{$service}++;
					}
				}
				if (/\}/) {
					undef $service;
					undef $IN_DEFAULTS_SECTION;
				}
			} elsif ($IN_SERVICE_SECTION) {
				$disabled{$service}++
					if ((/^[ \t]*disable\b/ && /\byes\b/)
					|| (/^[ \t]*flags\b/ && /\bDISABLE\b/));
				if (/\}/) {
					push @enabled, $service
						unless $disabled{$service};
					undef $service;
					undef $IN_SERVICE_SECTION;
				}
			} else {
				if (/^[ \t]*defaults\b/) {
					$IN_DEFAULTS_SECTION++;
				} elsif (/^[ \t]*service[ \t]+(\w+)\b/) {
					$service = $+;
					if ($ENABLED_EXISTS) {
						next unless $enabled{$service};
					}
					$IN_SERVICE_SECTION++;
				}
			}
		}
		close IN;
	}

	return @enabled;
}

# Returns list of enabled services.
sub getEnabled_src {
	my $IN_SERVICE_SECTION;
	my @enabled;

	open IN, "/bin/lssrc -l -s inetd |"
		or report("ERR", "Can't exec lssrc.");
	while (<IN>) {
		if ($IN_SERVICE_SECTION) {
			# Each line is one subserver; one space at beginning of
			# line, followed by subserver name, then other fields
			# separated by spaces.
			/^ (\w+)\s+/ and push @enabled, $+;
		} else {
			# Start of section listing inetd subservers looks like:
			# Service       Command         Description Status
			/^Service/ and $IN_SERVICE_SECTION++;
		}
	}
	close IN;

	return @enabled;
}

# Returns list of enabled services.
sub getEnabled_smf {
	my @enabled;

	open IN, "/usr/sbin/inetadm | /bin/tail +2 |"
		or report("ERR", "Can't exec inetadm.");
	while (<IN>) {
		# Skip unless enabled.
		/^enabled/ or next;
		# Pick out field containing name of service.
		my ($dum1, $dum2, $service) = split;
		# Abbreviate somewhat, by removing "svc:" at beginning...
		$service =~ s/^svc://;
		# ...and ":default" at end, if any...
		$service =~ s/:default$//;
		# ...and then removing 1st component of name, which is usually
		# uninteresting in that it doesn't help specify the service.
		($dum1, $dum2, $service) = split /\//, $service, 3;
		push @enabled, $service if defined $service;
	}
	close IN;

	return @enabled;
}

# Returns a list of enabled services.
sub getEnabled_launchd {
	my @enabled;

	open IN, "/bin/launchctl list |"
		or report("ERR", "Can't exec launchctl.");
	foreach my $service (<IN>) {
		chomp $service;
		report("DEBUG", "Checking $service...");
		my $svcFile = getConfig_launchd($service);
		unix("/usr/bin/grep -l \\<key\\>inetdCompatibility\\</key\\> $svcFile 2>&1 1>/dev/null") or next;
		# Strip out all but final segment of service name.
		$service =~ s/^.+\.//;
		# Strip off trailing "d", if there is one.
		$service =~ s/d$//;
		report("DEBUG", "$service is an inetd service.");
		push @enabled, $service;
	}
	close IN;

	return @enabled;
}

# Returns list consisting of a string indicating status of service, a Boolean
# indicating whether service is IPv6-enabled, a Boolean indicating whether
# service is wrapped with tcp_wrappers, and a list of files in which
# the service is defined.
sub getStatus {
	my $service = shift;
	my (%total, %enabled, %ip6, %wrapped);
	my ($total, $enabled, $ip6, $wrapped);
	my @matchedConfigFiles;
	my ($status, $IP6, $WRAPPED);

	foreach my $file (@configFiles) {
		# Get numbers for each config file.
		($total{$file}, $enabled{$file}, $ip6{$file}, $wrapped{$file})
			= countServices($service, $file);

		# Keep running totals.
		$total += $total{$file};
		$enabled += $enabled{$file};
		$ip6 += $ip6{$file};
		$wrapped += $wrapped{$file};

		# If the service defined in this file, keep track of it.
		$total{$file} and push @matchedConfigFiles, $file;
	}

	# Determine status.
	if ($total > 0) {
		$status = "multiply defined and " if ($total > 1);
		if ($enabled > 0) {
			$status .= "partially " if ($enabled != $total);
			$status .= "enabled";
			# Indicate IPv6 if any service has enabled it.
			$IP6++ if $ip6;
			# Indicate tcp_wrappers usage only if all enabled
			# services are wrapped.
			$WRAPPED++ if $wrapped == $enabled;
		} elsif ($enabled == 0) {
			$status .= "disabled";
		} else {
			$status = "undefined";
		}
	} elsif ($total == 0) {
		$status = "nonexistent";
	} else {
		$status = "undefined";
	}

	return ($status, $IP6, $WRAPPED, @matchedConfigFiles);
}

# Returns the number of service entries found in config files:
# total, enabled (uncommented), using IPv6, and wrapped.
sub countServices {
	my $service = shift;
	my $configFile = shift;
	my $DISABLED;
	my ($total, $enabled, $ip6, $wrapped) = (0, 0, 0, 0);

	open IN, "<$configFile" or report("ERR", "Can't open $configFile.");
	while (<IN>) {
		undef $DISABLED;
		s/^[ \t]*#[# \t]*// and $DISABLED++;
		my ($name, $dum1, $proto, $dum2, $dum3, $path) = split;
		if (defined $name and $name eq $service) {
			if ($os eq "solaris") {
				# hack to work around comment in stock Solaris
				# inetd.conf
				next if $proto eq "shell";
				$ip6++ if $proto =~ /6$/;
			}
			$total++;
			$enabled++ unless $DISABLED;
			$wrapped++ if $path =~ /\/tcpd$/;
		}
	}
	close IN;

	return ($total, $enabled, $ip6, $wrapped);
}

# Returns list consisting of a string indicating status of service, a Boolean
# indicating whether service is IPv6-enabled, the file in which the defaults
# section is located, and a list of files in which the service is defined.
#
# With xinetd, services are enabled by default if a service entry exists, or if
# it's an internal service. Conditions for disabling are:
#
#	existence of an "enabled" line in the "defaults" section, in which the
#		service is *not* listed
#		(Documentation indicates that "enabled" lines can exist in
#		individual service entries, but I'm ignoring that possibility
#		since it doesn't make much sense.)
#	existence of a "disabled" line in the "defaults" section, in which the
#		service *is* listed
#	existence of a "disable = yes" line in the service entry
#	setting of the "DISABLE" flag in a "flags" line in the service entry
#
# If any condition for disabling is found, the service is disabled. (I.e., the
# various "disable" methods override listing of a service in "enabled".)
sub getStatus_x {
	my $service = shift;
	my (%defaults, %total, %disabled, %ip6);
	my ($defaults, $total, $disabled, $ip6) = (0, 0, 0, 0);
	my @matchedConfigFiles;
	my ($status, $IP6, $defaultConfigFile);

	my $SERVICE_IS_INTERNAL;
	my @xinetdInternalService = (
		"chargen", "chargen-udp", "daytime", "daytime-udp",
		"echo", "echo-udp", "time", "time-udp",
		"discard", "servers", "services", "xadmin"
	);
	foreach (@xinetdInternalService) {
		/^$service$/ and $SERVICE_IS_INTERNAL++;
	}

	foreach my $file (@configFiles) {
		($defaults{$file}, $total{$file}, $disabled{$file}, $ip6{$file})
			= findService_x($service, $file);

		# Keep running totals.
		$defaults += $defaults{$file};
		$total += $total{$file};
		$disabled += $disabled{$file};
		$ip6 += $ip6{$file};

		# Error on multiple defaults sections, otherwise note location
		# of defaults section.
		$defaults > 1 and report("ERR", "Multiple files have 'defaults' section, including $file.");
		$defaults{$file} and $defaultConfigFile = $file;

		# If the service is defined in this file, or it has a "defaults"
		# section, pass it along for further use.
		($total{$file} || $defaults{$file})
			and push @matchedConfigFiles, $file;
	}

	# Internal services are enabled by default, even without service
	# definitions. Other services require service definitions.
	if (($total > 0) || ($SERVICE_IS_INTERNAL && ($total == 0))) {
		$status = "multiply defined and " if ($total > 1);
		if ($disabled > 0) {
			if ($disabled >= $total) {
				# The disabled count can be one higher than the
				# total number of service definitions, due to
				# the defaults section.
				$status .= "disabled";
			} else {
				# Can't be sure, because one disabling, in the
				# defaults section, can nullify any number of
				# service definitions.
				$status .= "possibly enabled";
				# Indicate IPv6 if any service has enabled it.
				$IP6++ if $ip6;
			}
		} elsif ($disabled == 0) {
			$status .= "enabled";
			$IP6++ if $ip6;
		} else {
			$status = "undefined";
		}
	} elsif (! $SERVICE_IS_INTERNAL && ($total == 0)) {
		$status = "nonexistent";
	} else {
		$status = "undefined";
	}

	# If no "defaults" section is found, use the main xinetd config file,
	# which should be first in the list.
	unless ($defaultConfigFile) {
		$defaultConfigFile = $configFiles[0];
		push @matchedConfigFiles, $defaultConfigFile;
	}

	return ($status, $IP6, $defaultConfigFile, @matchedConfigFiles);
}

# Returns a list consisting of numbers of entries found: defaults sections,
# service definitions, entries disabled, and entries using IPv6.
sub findService_x {
	my $service = shift;
	my $configFile = shift;
	my ($IN_SECTION, $DISABLED);
	my ($defaults, $total, $disabled, $ip6) = (0, 0, 0, 0);

	open IN, "<$configFile" or report("ERR", "Can't open $configFile.");
	while (<IN>) {
		if ($IN_SECTION) {
			$DISABLED++
				if ((/^[ \t]*enabled\b/ && ! /\b$service\b/)
				|| (/^[ \t]*disabled\b/ && /\b$service\b/)
				|| (/^[ \t]*disable\b/ && /\byes\b/)
				|| (/^[ \t]*flags\b/ && /\bDISABLE\b/));
			if (/\}/) {
				# Count only one disabling per section.
				undef $IN_SECTION;
				$disabled++ if $DISABLED;
			}
		} else {
			if (/^[ \t]*defaults\b/) {
				$defaults++;
				$IN_SECTION++;
			} elsif (/^[ \t]*service[ \t]+$service\b/) {
				$total++;
				$IN_SECTION++;
			}
		}
	}
	close IN;

	return ($defaults, $total, $disabled, $ip6);
}

# Returns list containing a string indicating status of service.
sub getStatus_src {
	my $service = shift;
	my $status;

	report("DEBUG", "Seeking status on $service.");
	my @statusLines = `/bin/lssrc -l -s inetd`;
	if (grep /^ $service\b.+\bactive$/, @statusLines) {
		$status = "enabled";
	} else {
		$status = "disabled or nonexistent";
      	}

	return ($status);
}

# Returns list consisting of a string indicating status of service, and a
# Boolean indicating whether service is wrapped with tcp_wrappers.
sub getStatus_smf {
	my $service = shift;
	my ($status, $WRAPPED);

	report("DEBUG", "Seeking status on $service.");
	$status = `/usr/sbin/inetadm | /bin/grep "$service"`;
	if ($status) {
		my ($enabled, $state) = split /\s+/, $status;
		$status = "$enabled";
		$status .= " and $state" unless $enabled eq "disabled";
		`/usr/sbin/inetadm -l "$service" | grep tcp_wrappers | grep TRUE` and $WRAPPED++;
	} else {
		$status = "nonexistent";
	}

	return ($status, $WRAPPED);
}

# Returns list containing a string indicating status of service.
sub getStatus_launchd {
	my $service = shift;
	my $status;

	report("DEBUG", "Seeking status on $service.");
	my @statusLines = `/bin/launchctl list`;
	if (grep /$service(d?)$/, @statusLines) {
		$status = "enabled";
	} else {
		$status = "disabled or nonexistent";
      	}

	return ($status);
}

# Returns path to plist for service.
sub getConfig_launchd {
	my $service = shift;
	my @configFiles = ();

	# Initial pass using simple grep.
	my @configFilesToTry = `/usr/bin/grep -l $service /System/Library/LaunchDaemons/* /Library/LaunchDaemons/* 2>/dev/null`;

	# I hope no one ever makes config filenames with spaces in them...
	foreach my $inFile (@configFilesToTry) {
		open IN, "<$inFile"
			or report("ERR", "Can't open $inFile for reading.");
		my @inLines = <IN>;
		close IN;
		grep /$service(d?)<\/string>$/, @inLines
			and push @configFiles, $inFile;
	}

	if (@configFiles < 1) {
		report("ERR", "Couldn't find config file for $service.");
	} elsif (@configFiles > 1) {
		report("ERR", "Found multiple config files for $service: @configFiles .");
	}

	my $configFile = shift @configFiles;
	chomp $configFile;
	report("DEBUG", "Service file is $configFile.");

	return $configFile;
}

sub updateConfig {
	my $inFile = shift;
	my $inPattern = shift;
	my $outPattern = shift;
	my $outFile = "$inFile.tmp";

	open IN, "<$inFile"
		or report("ERR", "Can't open $inFile for reading.");
	my @inLines = <IN>;
	close IN;

	my @outLines = map { s/$inPattern/$outPattern/; $_ } @inLines;

	generateConfig($outFile, @outLines);
	safeMove($outFile, $inFile);
}

sub updateConfig_x {
	my $inFile = shift;
	my $defaultConfigFile = shift;
	my $service = shift;
	my $changeFlag = shift;
	my @outLines;
	my $outFile = "$inFile.tmp";

	if ($changeFlag eq "DISABLE") {
		# If $service is defined, just want to append $service to
		# "disabled" line in "defaults" section.
		# If not defined, want to disable all services by replacing
		# "enabled" line in "defaults" section to have empty list.
		return unless $inFile eq $defaultConfigFile;
		@outLines = getDisableLines_x($inFile, $service);
	} elsif ($changeFlag eq "ENABLE") {
		# Add $service to "enabled" line, remove from "disabled" line,
		# remove "disable" line, remove "DISABLE" flag.
		@outLines = getEnableLines_x($inFile, $service);
		$service or report("ERR", "No service specified to enable.");
	} else {
		report("ERR", "Invalid change flag: $changeFlag");
	}

	generateConfig($outFile, @outLines);
	safeMove($outFile, $inFile);
}

# Takes list consisting of input filename and service to disable.
# Returns list containing transformed lines of the input file.
sub getDisableLines_x {
	my $inFile = shift;
	my $service = shift;
	my ($IN_SECTION, $DISABLED);
	my $outLine;
	my @outLines;

	if ($service) {
		$outLine = "\tdisabled\t\t= $service\n";
	} else {
		$outLine = "\tenabled\t\t\t=\n";
	}

	open IN, "$inFile" or report("ERR", "Can't open $inFile for reading.");
	while (<IN>) {
		if ($IN_SECTION) {
			if (/^[ \t]*disabled\b/ && $service) {
				# Append $service to "disabled" line.
				chomp;
				$_ .= " $service\n";
				$DISABLED++;
			} elsif (/^[ \t]*enabled\b/ && $service) {
				# Remove $service from "enabled" line. Not
				# strictly necessary, but keeps things clean.
				s/[ \t]+$service\b//g;
			} elsif (/^[ \t]*enabled\b/ && ! $service) {
				# If no service defined, disable all by
				# clearing out "enabled" line.
				$_ = $outLine;
				$DISABLED++;
			}
			if (/\}/) {
				if ($DISABLED) {
					# Multiple "(dis|en)abled" lines.
					$DISABLED > 1 and report("ERR", "Syntax error in 'defaults' section of $inFile.");
				} else {
					push @outLines, "$outLine";
					$DISABLED++;
				}
				undef $IN_SECTION;
			}
		} else {
			/^[ \t]*defaults\b/ and $IN_SECTION++;
		}
		push @outLines, $_;
	}
	# If no defaults section found.
	unless ($DISABLED) {
		unshift @outLines, "defaults\n", "{\n", "$outLine", "\n";
		$DISABLED++;
	}
	close IN;

	return @outLines;
}

# Takes list consisting of input filename and service to enable.
# Returns list containing transformed lines of the input file.
sub getEnableLines_x {
	my $inFile = shift;
	my $service = shift;
	my $IN_SECTION;
	my @outLines;

	open IN, "$inFile" or report("ERR", "Can't open $inFile for reading.");
	while (<IN>) {
		if ($IN_SECTION) {
			if (/^[ \t]*enabled\b/ && ! /\b$service\b/) {
				# Append $service to "enabled" line.
				chomp;
				$_ .= " $service\n";
			} elsif (/^[ \t]*disabled\b/) {
				# Remove $service from "disabled" line.
				s/[ \t]+$service\b//g;
				# If no values left, delete line.
				/=\s*$/ and $_ = "";
			} elsif (/^[ \t]*disable\b/) {
				# Clear out "disabled" line in service def.
				$_ = "";
			} elsif (/^[ \t]*flags\b/) {
				# Remove "DISABLE" from "flags" line in service
				# def.
				s/[ \t]+DISABLE\b//g;
				# If no values left, delete line.
				/=\s*$/ and $_ = "";
			}
			/\}/ and undef $IN_SECTION;
		} else {
			/^[ \t]*defaults\b/ and $IN_SECTION++;
			/^[ \t]*service[ \t]+$service\b/ and $IN_SECTION++;
		}
		push @outLines, $_;
	}
	close IN;

	return @outLines;
}

sub generateConfig {
	my $file = shift;
	my @lines = @_;

	unless (sysopen OUT, $file, O_WRONLY|O_EXCL|O_CREAT, 0644) {
		report("WARN", "$file locked. Wait...");
		sleep 3;
		sysopen OUT, $file, O_WRONLY|O_EXCL|O_CREAT, 0644
			or report("ERR", "Still locked. Try later, or remove $file.");
	}
	print OUT @lines;
	close OUT;
}

sub safeMove {
	my $fromFile = shift;
	my $toFile = shift;

	my $toDir = dirname $toFile;
	my $toBase = basename $toFile;
	my $bakFile = "$toDir/.$toBase.bak";

	-f $bakFile and (unlink $bakFile
		or report("WARN", "Can't remove $bakFile."));
	link $toFile, $bakFile
		or report("ERR", "Can't link $bakFile to $toFile.");
	rename $fromFile, $toFile
		or report("ERR", "Can't move $fromFile to $toFile.");
}

sub unix {
	# UNIX system calls generally return 0 for success, while Perl treats 0
	# as false. Need to return negation of system call return value.
	my $command = shift;
	return ! system $command;
}

sub report {
	my ($severity, $message) = @_;

	if	($severity eq "ERR")	{ die "$progname: $message\n"; }
	elsif	($severity eq "WARN")	{ warn "$progname: $message\n"; }
	elsif	($severity eq "INFO")	{ print "$progname: $message\n"; }
	elsif	($severity eq "DEBUG")	{ warn "$progname: DEBUG: $message\n"
						if $DEBUG; }
	else	{ die "$progname: Undefined error. $message\n"; }
}
